//----------------------------------------------------------------------------------
// XbhNX
// Copyright (c) 2008  N099
//----------------------------------------------------------------------------------
#include <windows.h>
#include <process.h>

#include "thrproc.h"

//constructor
ThrProc::ThrProc() : m_idThr(0), m_priThr(0), m_param(NULL), m_func(NULL) 
{
	m_hReq = m_hFin = m_hThr = NULL;
	m_stop = m_req = false;
	InitializeCriticalSection(&m_cs);
}

ThrProc::~ThrProc()
{
	Shutdown();
	CloseEvent();
	DeleteCriticalSection(&m_cs);
}

int ThrProc::Init(THRPROCFUNC fn, void *param)
{
	EnterCriticalSection(&m_cs);

	//Xbh쐬ς
	if(m_hThr){
		LeaveCriticalSection(&m_cs);
		return 0;
	}

	m_func = fn;
	m_param = param;
	m_stop = false;

	//Cxg̍쐬
	if(!m_hReq){
		m_hReq = CreateEvent(NULL, FALSE, FALSE, NULL);
	}
	
	if(!m_hFin){
		m_hFin = CreateEvent(NULL, FALSE, FALSE, NULL);
	}

	if(!m_hReq || !m_hFin){
		LeaveCriticalSection(&m_cs);
		return -1;
	}

	ResetEvent(m_hReq);
	ResetEvent(m_hFin);

	//Xbh쐬
	m_hThr = reinterpret_cast<HANDLE>(_beginthreadex(NULL, 0, IniProc, this, 0, &m_idThr));
	if(!m_hThr){
		LeaveCriticalSection(&m_cs);
		return -1;
	}

	//Dxݒ
	m_priThr = GetThreadPriority(GetCurrentThread()) + 1;
	SetThreadPriority(m_hThr, m_priThr);
	
	LeaveCriticalSection(&m_cs);
	return 0;
}

int ThrProc::Start()
{
	if(!m_hThr || m_stop) return -1;

	//Dxグ
	int pri = GetThreadPriority(GetCurrentThread()) + 1;
	if(pri != m_priThr){
		m_priThr = pri;
		SetThreadPriority(m_hThr, m_priThr);
	}

	m_req = true;
	SetEvent(m_hReq);

	return 0;
}

int ThrProc::Wait()
{
	if(!m_hThr || m_stop) return -1;

	m_req = false;
	if(WAIT_OBJECT_0 != WaitForSingleObject(m_hFin, INFINITE)){
		return -1;
	}

	return 0;
}

int ThrProc::Shutdown()
{
	EnterCriticalSection(&m_cs);

	HANDLE hThr = InterlockedExchangePointer(&m_hThr, NULL);
	if(hThr == NULL){
		LeaveCriticalSection(&m_cs);
		return 0;
	}

	//Xbh~
	m_stop = true;
	SetEvent(m_hReq);
	WaitForSingleObject(hThr, INFINITE);

	LeaveCriticalSection(&m_cs);

	return 0;
}

//private
void ThrProc::CloseEvent()
{
	EnterCriticalSection(&m_cs);
	
	if(m_hReq){
		CloseHandle(m_hReq);
		m_hReq = NULL;
	}

	if(m_hFin){
		CloseHandle(m_hFin);
		m_hFin = NULL;
	}

	LeaveCriticalSection(&m_cs);
}

void ThrProc::Proc()
{
	while(!m_stop){
		if(WAIT_OBJECT_0 != WaitForSingleObject(m_hReq, INFINITE)){
			m_stop = true;
		}

		if(!m_stop && m_func){
			m_func(m_param);
		}

		SetEvent(m_hFin);
	}

	SetEvent(m_hFin);
}

unsigned int __stdcall ThrProc::IniProc(void *param)
{
	static_cast<ThrProc*>(param)->Proc();

	_endthreadex(0);
	return 0;
}

